(***********************************************************************)
(*                                                                     *)
(*                               TAToo                                 *)
(*                                                                     *)
(*                     Kim Nguyen, LRI UMR8623                         *)
(*                   Université Paris-Sud & CNRS                       *)
(*                                                                     *)
(*  Copyright 2010-2013 Université Paris-Sud and Centre National de la *)
(*  Recherche Scientifique. All rights reserved.  This file is         *)
(*  distributed under the terms of the GNU Lesser General Public       *)
(*  License, with the special exception on linking described in file   *)
(*  ../LICENSE.                                                        *)
(*                                                                     *)
(***********************************************************************)

(*
  Time-stamp: <Last modified on 2013-04-04 21:16:04 CEST by Kim Nguyen>
*)

INCLUDE "utils.ml"
open Format

type predicate = | First_child
                 | Next_sibling
                 | Parent
                 | Previous_sibling
                 | Stay
                 | Is_first_child
                 | Is_next_sibling
                 | Is of (Tree.NodeKind.t)
                 | Has_first_child
                 | Has_next_sibling

let is_move p = match p with
| First_child | Next_sibling
| Parent | Previous_sibling | Stay -> true
| _ -> false


type atom = predicate * bool * State.t

module Atom : (Formula.ATOM with type data = atom) =
struct

  module Node =
  struct
    type t = atom
    let equal n1 n2 = n1 = n2
    let hash n = Hashtbl.hash n
  end

  include Hcons.Make(Node)

  let print ppf a =
    let p, b, q = a.node in
    if not b then fprintf ppf "%s" Pretty.lnot;
    match p with
    | First_child -> fprintf ppf "FC(%a)" State.print q
    | Next_sibling -> fprintf ppf "NS(%a)" State.print q
    | Parent -> fprintf ppf "FC%s(%a)" Pretty.inverse State.print q
    | Previous_sibling -> fprintf ppf "NS%s(%a)" Pretty.inverse State.print q
    | Stay -> fprintf ppf "%s(%a)" Pretty.epsilon State.print q
    | Is_first_child -> fprintf ppf "FC%s?" Pretty.inverse
    | Is_next_sibling -> fprintf ppf "NS%s?" Pretty.inverse
    | Is k -> fprintf ppf "is-%a?" Tree.NodeKind.print k
    | Has_first_child -> fprintf ppf "FC?"
    | Has_next_sibling -> fprintf ppf "NS?"

  let neg a =
    let p, b, q = a.node in
    make (p, not b, q)


end

module SFormula =
struct
  include Formula.Make(Atom)
  open Tree.NodeKind
  let mk_atom a b c = atom_ (Atom.make (a,b,c))
  let mk_kind k = mk_atom (Is k) true State.dummy
  let has_first_child =
    (mk_atom Has_first_child true State.dummy)

  let has_next_sibling =
    (mk_atom Has_next_sibling true State.dummy)

  let is_first_child =
    (mk_atom Is_first_child true State.dummy)

  let is_next_sibling =
    (mk_atom Is_next_sibling true State.dummy)

  let is_attribute =
    (mk_atom (Is Attribute) true State.dummy)

  let is_element =
    (mk_atom (Is Element) true State.dummy)

  let is_processing_instruction =
    (mk_atom (Is ProcessingInstruction) true State.dummy)

  let is_comment =
    (mk_atom (Is Comment) true State.dummy)

  let first_child q =
  and_
    (mk_atom First_child true q)
    has_first_child

  let next_sibling q =
  and_
    (mk_atom Next_sibling true q)
    has_next_sibling

  let parent q =
  and_
    (mk_atom Parent true q)
    is_first_child

  let previous_sibling q =
  and_
    (mk_atom Previous_sibling true q)
    is_next_sibling

  let stay q =
    (mk_atom Stay true q)

  let get_states phi =
    fold (fun phi acc ->
      match expr phi with
      | Formula.Atom a -> let _, _, q = Atom.node a in
                          if q != State.dummy then StateSet.add q acc else acc
      | _ -> acc
    ) phi StateSet.empty

end


module Transition = Hcons.Make (struct
  type t = State.t * QNameSet.t * SFormula.t
  let equal (a, b, c) (d, e, f) =
    a == d && b == e && c == f
  let hash (a, b, c) =
    HASHINT4 (PRIME1, a, ((QNameSet.uid b) :> int), ((SFormula.uid c) :> int))
end)


module TransList : sig
  include Hlist.S with type elt = Transition.t
  val print : Format.formatter -> ?sep:string -> t -> unit
end =
  struct
    include Hlist.Make(Transition)
    let print ppf ?(sep="\n") l =
      iter (fun t ->
        let q, lab, f = Transition.node t in
        fprintf ppf "%a, %a -> %a%s" State.print q QNameSet.print lab SFormula.print f sep) l
  end


type t = {
  id : Uid.t;
  mutable states : StateSet.t;
  mutable selection_states: StateSet.t;
  transitions: (State.t, (QNameSet.t*SFormula.t) list) Hashtbl.t;
  mutable cache2 : TransList.t Cache.N2.t;
  mutable cache6 : (TransList.t*StateSet.t) Cache.N6.t;
}

let next = Uid.make_maker ()

let dummy2 = TransList.cons
  (Transition.make (State.dummy,QNameSet.empty, SFormula.false_))
  TransList.nil

let dummy6 = (dummy2, StateSet.empty)


let create s ss =
  let auto = { id = next ();
               states = s;
               selection_states = ss;
               transitions = Hashtbl.create 17;
               cache2 = Cache.N2.create dummy2;
               cache6 = Cache.N6.create dummy6;
             }
  in
  at_exit (fun () ->
    let n6 = ref 0 in
    let n2 = ref 0 in
    Cache.N2.iteri (fun _ _ _ b -> if b then incr n2) auto.cache2;
    Cache.N6.iteri (fun _ _ _ _ _ _ _ b -> if b then incr n6) auto.cache6;
    Format.eprintf "INFO: automaton %i, cache2: %i entries, cache6: %i entries\n%!"
      (auto.id :> int) !n2 !n6;
    let c2l, c2u = Cache.N2.stats auto.cache2 in
    let c6l, c6u = Cache.N6.stats auto.cache6 in
    Format.eprintf "INFO: cache2: length: %i, used: %i, occupation: %f\n%!" c2l c2u (float c2u /. float c2l);
    Format.eprintf "INFO: cache6: length: %i, used: %i, occupation: %f\n%!" c6l c6u (float c6u /. float c6l)

  );
  auto

let reset a =
  a.cache2 <- Cache.N2.create dummy2;
  a.cache6 <- Cache.N6.create dummy6


let get_trans_aux a tag states =
  StateSet.fold (fun q acc0 ->
    try
      let trs = Hashtbl.find a.transitions q in
      List.fold_left (fun acc1 (labs, phi) ->
        if QNameSet.mem tag labs then TransList.cons (Transition.make (q, labs, phi)) acc1 else acc1) acc0 trs
    with Not_found -> acc0
  ) states TransList.nil


let get_trans a tag states =
  let trs =
    Cache.N2.find a.cache2
      (tag.QName.id :> int) (states.StateSet.id :> int)
  in
  if trs == dummy2 then
    let trs = get_trans_aux a tag states in
    (Cache.N2.add
       a.cache2
       (tag.QName.id :> int)
       (states.StateSet.id :> int) trs; trs)
  else trs



let eval_form phi fcs nss ps ss is_left is_right has_left has_right kind =
  let rec loop phi =
    begin match SFormula.expr phi with
      Formula.True | Formula.False -> phi
    | Formula.Atom a ->
        let p, b, q = Atom.node a in begin
          match p with
          | First_child ->
              if b == StateSet.mem q fcs then SFormula.true_ else phi
          | Next_sibling ->
              if b == StateSet.mem q nss then SFormula.true_ else phi
          | Parent | Previous_sibling ->
              if b == StateSet.mem q ps then SFormula.true_ else phi
          | Stay ->
              if b == StateSet.mem q ss then SFormula.true_ else phi
          | Is_first_child -> SFormula.of_bool (b == is_left)
          | Is_next_sibling -> SFormula.of_bool (b == is_right)
          | Is k -> SFormula.of_bool (b == (k == kind))
          | Has_first_child -> SFormula.of_bool (b == has_left)
          | Has_next_sibling -> SFormula.of_bool (b == has_right)
        end
    | Formula.And(phi1, phi2) -> SFormula.and_ (loop phi1) (loop phi2)
    | Formula.Or (phi1, phi2) -> SFormula.or_  (loop phi1) (loop phi2)
    end
  in
  loop phi

let int_of_conf is_left is_right has_left has_right kind =
  ((Obj.magic kind) lsl 4) lor
    ((Obj.magic is_left) lsl 3) lor
    ((Obj.magic is_right) lsl 2) lor
    ((Obj.magic has_left) lsl 1) lor
    (Obj.magic has_right)

let eval_trans auto ltrs fcs nss ps ss is_left is_right has_left has_right kind =
  let n = int_of_conf is_left is_right has_left has_right kind
  and k = (fcs.StateSet.id :> int)
  and l = (nss.StateSet.id :> int)
  and m = (ps.StateSet.id :> int) in
  let rec loop ltrs ss =
    let i = (ltrs.TransList.id :> int)
    and j = (ss.StateSet.id :> int) in
    let (new_ltrs, new_ss) as res =
      let res = Cache.N6.find auto.cache6 i j k l m n in
      if res == dummy6 then
        let res =
          TransList.fold (fun trs (acct, accs) ->
            let q, lab, phi = Transition.node trs in
            if StateSet.mem q accs then (acct, accs) else
              let new_phi =
                eval_form
                  phi fcs nss ps accs
                  is_left is_right has_left has_right kind
              in
              if SFormula.is_true new_phi then
                (acct, StateSet.add q accs)
              else if SFormula.is_false new_phi then
                (acct, accs)
              else
                let new_tr = Transition.make (q, lab, new_phi) in
                (TransList.cons new_tr acct, accs)
          ) ltrs (TransList.nil, ss)
        in
        Cache.N6.add auto.cache6 i j k l m n res; res
      else
        res
    in
    if new_ss == ss then res else
      loop new_ltrs new_ss
  in
  loop ltrs ss





(*
  [add_trans a q labels f] adds a transition [(q,labels) -> f] to the
  automaton [a] but ensures that transitions remains pairwise disjoint
*)

let add_trans a q s f =
  let trs = try Hashtbl.find a.transitions q with Not_found -> [] in
  let cup, ntrs =
    List.fold_left (fun (acup, atrs) (labs, phi) ->
      let lab1 = QNameSet.inter labs s in
      let lab2 = QNameSet.diff labs s in
      let tr1 =
        if QNameSet.is_empty lab1 then []
        else [ (lab1, SFormula.or_ phi f) ]
      in
      let tr2 =
        if QNameSet.is_empty lab2 then []
        else [ (lab2, SFormula.or_ phi f) ]
      in
      (QNameSet.union acup labs, tr1@ tr2 @ atrs)
    ) (QNameSet.empty, []) trs
  in
  let rem = QNameSet.diff s cup in
  let ntrs = if QNameSet.is_empty rem then ntrs
    else (rem, f) :: ntrs
  in
  Hashtbl.replace a.transitions q ntrs

let _pr_buff = Buffer.create 50
let _str_fmt = formatter_of_buffer _pr_buff
let _flush_str_fmt () = pp_print_flush _str_fmt ();
  let s = Buffer.contents _pr_buff in
  Buffer.clear _pr_buff; s

let print fmt a =
  fprintf fmt
    "\nInternal UID: %i@\n\
     States: %a@\n\
     Selection states: %a@\n\
     Alternating transitions:@\n"
    (a.id :> int)
    StateSet.print a.states
    StateSet.print a.selection_states;
  let trs =
    Hashtbl.fold
      (fun q t acc -> List.fold_left (fun acc (s , f) -> (q,s,f)::acc) acc t)
      a.transitions
      []
  in
  let sorted_trs = List.stable_sort (fun (q1, s1, _) (q2, s2, _) ->
    let c = State.compare q1 q2 in - (if c == 0 then QNameSet.compare s1 s2 else c))
    trs
  in
  let _ = _flush_str_fmt () in
  let strs_strings, max_pre, max_all = List.fold_left (fun (accl, accp, acca) (q, s, f) ->
    let s1 = State.print _str_fmt q; _flush_str_fmt () in
    let s2 = QNameSet.print _str_fmt s;  _flush_str_fmt () in
    let s3 = SFormula.print _str_fmt f;  _flush_str_fmt () in
    let pre = Pretty.length s1 + Pretty.length s2 in
    let all = Pretty.length s3 in
    ( (q, s1, s2, s3) :: accl, max accp pre, max acca all)
  ) ([], 0, 0) sorted_trs
  in
  let line = Pretty.line (max_all + max_pre + 6) in
  let prev_q = ref State.dummy in
  List.iter (fun (q, s1, s2, s3) ->
    if !prev_q != q && !prev_q != State.dummy then fprintf fmt " %s\n%!"  line;
    prev_q := q;
    fprintf fmt " %s, %s" s1 s2;
    fprintf fmt "%s" (Pretty.padding (max_pre - Pretty.length s1 - Pretty.length s2));
    fprintf fmt " %s  %s@\n%!" Pretty.right_arrow s3;
  ) strs_strings;
  fprintf fmt " %s\n%!" line

(*
  [complete transitions a] ensures that for each state q
  and each symbols s in the alphabet, a transition q, s exists.
  (adding q, s -> F when necessary).
*)

let complete_transitions a =
  StateSet.iter (fun q ->
    let qtrans = Hashtbl.find a.transitions q in
    let rem =
      List.fold_left (fun rem (labels, _) ->
        QNameSet.diff rem labels) QNameSet.any qtrans
    in
    let nqtrans =
      if QNameSet.is_empty rem then qtrans
      else
        (rem, SFormula.false_) :: qtrans
    in
    Hashtbl.replace a.transitions q nqtrans
  ) a.states

let cleanup_states a =
  let memo = ref StateSet.empty in
  let rec loop q =
    if not (StateSet.mem q !memo) then begin
      memo := StateSet.add q !memo;
      let trs = try Hashtbl.find a.transitions q with Not_found -> [] in
      List.iter (fun (_, phi) ->
        StateSet.iter loop (SFormula.get_states phi)) trs
    end
  in
  StateSet.iter loop a.selection_states;
  let unused = StateSet.diff a.states !memo in
  eprintf "Unused states %a\n%!" StateSet.print unused;
  StateSet.iter (fun q -> Hashtbl.remove a.transitions q) unused;
  a.states <- !memo

(* [normalize_negations a] removes negative atoms in the formula
   complementing the sub-automaton in the negative states.
   [TODO check the meaning of negative upward arrows]
*)

let normalize_negations auto =
  eprintf "Automaton before normalize_trans:\n";
  print err_formatter auto;
  eprintf "--------------------\n%!";

  let memo_state = Hashtbl.create 17 in
  let todo = Queue.create () in
  let rec flip b f =
    match SFormula.expr f with
      Formula.True | Formula.False -> if b then f else SFormula.not_ f
    | Formula.Or(f1, f2) -> (if b then SFormula.or_ else SFormula.and_)(flip b f1) (flip b f2)
    | Formula.And(f1, f2) -> (if b then SFormula.and_ else SFormula.or_)(flip b f1) (flip b f2)
    | Formula.Atom(a) -> begin
      let l, b', q = Atom.node a in
      if q == State.dummy then if b then f else SFormula.not_ f
      else
        if b == b' then begin
        (* a appears positively, either no negation or double negation *)
          if not (Hashtbl.mem memo_state (q,b)) then Queue.add (q,true) todo;
          SFormula.atom_ (Atom.make (l, true, q))
        end else begin
        (* need to reverse the atom
           either we have a positive state deep below a negation
           or we have a negative state in a positive formula
           b' = sign of the state
           b = sign of the enclosing formula
        *)
        let not_q =
          try
            (* does the inverted state of q exist ? *)
            Hashtbl.find memo_state (q, false)
          with
            Not_found ->
              (* create a new state and add it to the todo queue *)
              let nq = State.make () in
              auto.states <- StateSet.add nq auto.states;
              Hashtbl.add memo_state (q, false) nq;
              Queue.add (q, false) todo; nq
        in
        SFormula.atom_ (Atom.make (l, true, not_q))
      end
    end
  in
  (* states that are not reachable from a selection stat are not interesting *)
  StateSet.iter (fun q -> Queue.add (q, true) todo) auto.selection_states;

  while not (Queue.is_empty todo) do
    let (q, b) as key = Queue.pop todo in
    let q' =
      try
        Hashtbl.find memo_state key
      with
        Not_found ->
          let nq = if b then q else
              let nq = State.make () in
              auto.states <- StateSet.add nq auto.states;
              nq
          in
          Hashtbl.add memo_state key nq; nq
    in
    let trans = Hashtbl.find auto.transitions q in
    let trans' = List.map (fun (lab, f) -> lab, flip b f) trans in
    Hashtbl.replace auto.transitions q' trans';
  done;
  cleanup_states auto


